import sys

from . import shapes
from .scan_styles import ScanStyles
from .fib import FIBG2
from .resources import plot_stream_file, print_streamfile_time

REQUIRED_MAJOR = 3
REQUIRED_MINOR = 11  # Change this to your minimum required version

if sys.version_info < (REQUIRED_MAJOR, REQUIRED_MINOR):
    raise RuntimeError(
        f"The deFIBrillator package requires Python {REQUIRED_MAJOR}.{REQUIRED_MINOR} or higher."
        f"You are currently using Python {sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}."
    )

__all__ = ["shapes", "FIBG2", "plot_stream_file", "print_streamfile_time", "ScanStyles"]
